/******************************************************************************
 *
 *   FILE SPECIFICATION : Dltest.cpp
 *       TARGET MACHINE : PC / WINDOWS WORKSTATION
 *          ENVIRONMENT : Microsoft (c) Windows/NT
 *               SYSTEM : N/A
 *              PROJECT : N/A
 *
 *   COPYRIGHT
 *       Copyright 1995 - 1999 Ian Duff - Duff Software & Services
 *       ALL RIGHTS RESERVED
 *
 ************* F U N C T I O N A L    D E S C R I P T I O N *******************
 *
 *  Purpose:	This module demonstrates the functionality of the methods
 *				and properties of the CLinkedList class.
 *
 * Notes:
 *				This demonstration has been specifically designed for an MFC
 *				environment.
 *
 *	Member Function			Purpose
 *	------------------------------------------------------------------------
 *	dladd();		Add a node to the list.
 *	
 *
 ************* C O N F I G U R A T I O N   M A N A G E M E N T ****************
 *
 * @(#)$Date:		$
 * @(#)$Revision:	$
 * @(#)$Locker:		$
 * @(#)$State:		$
 *
 ************* M O D I F I C A T I O N   H I S T O R Y ************************
 *     Date       | Author        | Version  | Reason for change
 *----------------+---------------+----------+---------------------------------
 * 21-March-1999  | I.  Duff      | 2.0W     | Initial Version.
 *----------------+---------------+----------+---------------------------------
 *****************************************************************************/
// DlTest.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "DlTest.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// The one and only application object

CWinApp theApp;

using namespace std;

/* ---------------- module include files ---------------------------------- */

#include <conio.h>			// For getch()
#include "LinkedList.h"		// ClinkedList class definitions.

/* ---------------- module constant definitions --------------------------- */

const int	kMaxNodes = 10;		// Number of nodes for this demonstration.
const int	kSmallBuf = 80;		// Scratch buffer sizes.

/* ---------------- module typedef statements ----------------------------- */

typedef struct {					// This can be any user defined structure.
	char Data[ kSmallBuf ];
	char *myChar;					// Used to illustrate the FreeData function.
}
Structure;
/* ---------------- module functions begin here --------------------------- */

/******************************************************************************
 *  Function Name: Compare1()
 *
 *  Purpose:This is a user defined comparison function used by the linked list 
 *			functions dlfind(), dlqsort() and dlsort().
 * 
 *  Parameters:    Two structures to be compared.
 *
 *  Side Effects:  None.
 *
 *  Return Values: The result of strcmp().
 *
 *  Notes:
 *		1.	In this case, the "key" used is the char "Data" field of the
 *			user's STRUCTURE.
 *		2.	This function compares d1 against d2.(NORMAL COMPARE)
 *		3.	Used to test the comparison "change" function dlsetcompare();
 *		   
 ******************************************************************************/
int Compare1( const void* d1, const void *d2)
{
	Structure	*s1 = (Structure *)d1;
	Structure	*s2 = (Structure *)d2;

	return strcmp((char *)s1->Data, (char *)s2->Data );
}

/******************************************************************************
 *  Function Name: Compare2()
 *
 *  Purpose:This is a user defined comparison function used by the linked list 
 *			functions dlfind(), dlqsort() and dlsort().
 * 
 *  Parameters:    Two structures to be compared.
 *
 *  Side Effects:  None.
 *
 *  Return Values: The result of strcmp().
 *
 *  Notes:	
 *		1.	In this case, the "key" used is the char "Data" field of the
 *			user's STRUCTURE.
 *		2.	This function compares d2 against d1. (REVERSE COMPARE )
 *		3.	Used to test the comparison "change" function dlsetcompare();
 ******************************************************************************/
int Compare2( const void *d1, const void *d2)
{
	Structure	*s1 = (Structure *)d2;
	Structure	*s2 = (Structure *)d1;

	return strcmp( (char *)s1->Data, (char *)s2->Data );
}


/******************************************************************************
 *  Function Name: FreeData()
 *
 *  Purpose:	
 *		This user defined function removes any other "items" connected to the 
 *		linked list. For example, the structure defined may have other linked 
 *		lists or "allocated" nodes that need to be freed before the actual 
 *		linked list is removed from memory.  In the example STRUCTURE, a char 
 *		pointer myChar  has had memory allocated and a value inserted. This 
 *		function frees the previously allocated space back to memory as part 
 *		of the list's destructor, just prior to deleting the linked list.          
 *
 *  Parameters: The incoming void pointer which is currently pointing at the
 *		user defined STRUCTURE structure. 
 *
 *  Side Effects:  Memory is released back to the system.
 *
 *  Return Values: None (in this case).
 *
 *  Notes:
 *		1.	In this case, the "key" used is the char "Data" field 
 *			of the user's STRUCTURE.
 *		   
 ******************************************************************************/
void FreeData( LPVOID p )
{
	Structure	*pStructure = (Structure *)p;
	ASSERT( p );
	cout << "FreeData: deleting " << pStructure->myChar << endl;
	delete[] pStructure->myChar;
}



/******************************************************************************
 *  Function Name: DemonstrateLinkedLists()
 *
 *  Purpose:	
 *		Demonstrates just about all of the CLinkedList class member functions. 
 *
 *  Parameters: None.
 *
 *  Side Effects:  None.
 *
 *
 *  Notes:
 *		1.	My apologies to those who know that this function is too long. I
 *			decided to include it at the last minute, so that you can see for
 *			yourself that the functionality does indeed work.
 *			
 *		   
 ******************************************************************************/
int DemonstrateLinkedLists( void )
{
	// Greet the user first
	CString strHello;
	strHello.LoadString( IDS_HELLO );
	cout << (LPCTSTR)strHello << endl;

	Structure	myStructure;			// A user defined structure.
	Structure	*pStructure;			// Pointer to a user defined structure.
	int			retCode = -1;

	CLinkedList  myList( sizeof( Structure ), Compare1, FreeData );
	/*-----Examples of alternate calling methods -----------
	 * LinkedList	myList( sizeof( Structure ));
	 * LinkedList	myList( sizeof( Structure ), Compare2);
	 *------------------------------------------------------*/
	cout << "List Initialisation status = " 
		 << ( myList.GetStatus() == 0 ? "Successful" : "Fail");



	/**************************************************************************
	 *  Add some test elements to the list          
	 *	Start from high numbers to low ones.	
	 *************************************************************************/
	cout << endl << "List Successfully created" << endl;
	cout << "ADDING (" << kMaxNodes << ") nodes in descending order with insertion sort"<< endl;

	for ( int i = kMaxNodes; i > 0; i-- ) {
		sprintf( myStructure.Data,"DATA=%02d",i );
		// allocate some memory for the char pointer
		myStructure.myChar = new char[ kSmallBuf ];
		ASSERT( myStructure.myChar );
		// print a number into the new space
		sprintf( myStructure.myChar,"%02d", i );
		cout	<< "Adding Data=>" << myStructure.Data << "< myChar=" 
				<< myStructure.myChar << endl;
		// Add the new node to the list
		if (( retCode = myList.dladdins( ( LPVOID )&myStructure )) != 0 ) {
			cout << "Error " << retCode << " Adding to list" << endl;
			return ( retCode );
		}
	}
	cout << "Notice that we have inserted nodes in descending order." << endl;
	cout << "Because we have specified dladdins, we have forced the" << endl;
	cout << "nodes in the list to be sorted on insert" << endl;
	cout << "Press any key to continue";
	getch();

	/*-------------------------------------------------------
	 * Now traverse the list from start to end and print 
	 * out the contents of each structure in the list.  
	 *------------------------------------------------------*/
	cout << endl << "Printing list nodes from BEGIN to END of list" << endl;
	pStructure = (Structure *)myList.dltofirst();

	while ( pStructure ) {
		cout	<< ">" << pStructure->Data << "< myChar=" << pStructure->myChar 
				<< " Position=(" << myList.dlposn() << ")" << endl;
		pStructure = ( Structure *)myList.dlgofwd();
	}
	cout << "Finished - About to traverse backwards... " << endl;
	cout << "Press any key to continue";
	getch();
	cout << endl;

	/*-------------------------------------------------------
	 *      Now do the same from end to beginning of list.  
	 *------------------------------------------------------*/

	pStructure = ( Structure *)myList.dltolast();
	while ( pStructure ) 	{
		cout	<< ">" << pStructure->Data << "< myChar=" << pStructure->myChar 
				<< " Position=(" << myList.dlposn() << ")" << endl;
		pStructure = (Structure *)myList.dlgoback();
	}

	/*-------------------------------------------------------
	 * Now traverse the list from start to end AGAIN and print 
	 * out the contents of each structure in the list.  
	 * This time we use some of the overloaded operators.
	 *------------------------------------------------------*/
	cout << endl << "Printing list nodes AGAIN from BEGIN to END of list" << endl;
	cout << "This time using overloaded operators to achieve the same result" << endl;

	pStructure = ( Structure *)myList[ 0 ];
	while ( pStructure ) {
		cout	<< ">" << pStructure->Data << "< myChar=" << pStructure->myChar 
				<< " Position=(" << myList.dlposn() << ")" << endl;		
		pStructure = ( Structure *)myList++;
	}

	cout << "Finished - About to traverse backwards... " << endl;
	cout << "There are currently " << myList.dlcount() << " nodes in the list" << endl;
	cout << "Press any key to continue";
	getch();
	cout << endl;

	pStructure = ( Structure *)myList[ myList.dlcount() - 1 ];
	while ( pStructure ) {
		cout	<< ">" << pStructure->Data << "< myChar=" << pStructure->myChar 
				<< " Position=(" << myList.dlposn() << ")" << endl;		
		pStructure = ( Structure *)myList--;
	}

	/*-------------------------------------------------------
	 * Now we search for a particular node in the list. A   
	 * NULL pointer is returned ( by CLinkedList) if we can't find it
	 *------------------------------------------------------*/
	cout << "Now demonstrating the find functionality of the CLinkedList" << endl;
	cout << "This is effectively a sequential search of the CLinkedList class" << endl;
	cout << "Looking for DATA=05.......(Any key to continue)";
	getch();
	cout << endl;
	strcpy(myStructure.Data,"DATA=05");

	/*-------------------------------------------------------
	 * First we use the dlfind() method
	 *------------------------------------------------------*/

	if ( pStructure = (Structure *)myList.dlfind( &myStructure )) 
			cout << "Found >" << pStructure->Data << "< myChar=" << pStructure->myChar 
				 << " Position=(" << myList.dlposn() << ")" << endl;		
	else    cout << "Couldn't find it!" << endl;

	cout << "We may also use a binary search (obviously quicker) on a sorted list" << endl;
	cout << "In order to find a node.  This time, we use the dlbsearch() method" << endl;
	cout << "Looking for DATA=06.......(Any key to continue)";
	getch();
	cout << endl;

	/*-------------------------------------------------------
	 * Now we use the dlbsearch() method (binary search)
	 *------------------------------------------------------*/
	strcpy(myStructure.Data,"DATA=06");	// Look for a different one this time.
	if ( pStructure = ( Structure *)myList.dlbsearch( (LPVOID)&myStructure ))
			cout << "Found >" << pStructure->Data << "< myChar=" << pStructure->myChar 
				 << " Position=(" << myList.dlposn() << ")" << endl;		
	else    cout << "Couldn't find it!" << endl;

	// Add another node to the list
	strcpy(myStructure.Data,"DATA=00");	// Look for a different one this time.
	myStructure.myChar = new char[ kSmallBuf ];
	ASSERT( myStructure.myChar );
	// print a number into the new space
	sprintf( myStructure.myChar,"%02d", 0 );

	// Add it to the list (Use a different method)
	cout	<< "Adding another node to the list" << endl;
	cout	<< "Adding Data=>" << myStructure.Data << "< myChar=" 
			<< myStructure.myChar << endl;

	myList << (LPVOID)&myStructure;

	cout << "There are currently " << myList.dlcount() << " nodes in the list" << endl;
	pStructure = (Structure *)myList.dltofirst();
	while ( pStructure ) {
		cout	<< ">" << pStructure->Data << "< myChar=" << pStructure->myChar 
				<< " Position=(" << myList.dlposn() << ")" << endl;
		pStructure = ( Structure *)myList.dlgofwd();
	}

	cout << "When adding to the list, using the dladd() method appends to the " << endl;
	cout << "end of the list" << endl;
	cout << "The List is now NOT correctly sorted, so we try a dlbsearch() on the" << endl;
	cout << "list.  We should now expect a FAIL" << endl;

	/*----------------------------------------------------------------------
	 * Now we use the dlbsearch() method (binary search) expecting a fail
	 *----------------------------------------------------------------------*/
	strcpy(myStructure.Data,"DATA=06");	// Look for a different one this time.
	if ( pStructure = ( Structure *)myList.dlbsearch( (LPVOID)&myStructure ))
			cout << "Found >" << pStructure->Data << "< myChar=" << pStructure->myChar 
				 << " Position=(" << myList.dlposn() << ")" << endl;		
	else    {
		cout << "Couldn't find it! - Status = " << myList.GetStatus() << endl;
		cout << "This is the error number ListHasNotBeenSorted" << endl;
	}

	cout << "We now sort the list with dlqsort() and try again" << endl;
	cout << "Any key to continue" << endl;
	getch();
	myList.dlqsort();
	cout << endl;

	cout << "List has now been sorted by the qsort() method" << endl;
	cout << "Looking again for DATA=06.......(Any key to continue)";
	getch();
	cout << endl;

	/*----------------------------------------------------------------------
	 * Now we use the dlbsearch() method (binary search) expecting success
	 *----------------------------------------------------------------------*/
	if ( pStructure = ( Structure *)myList.dlbsearch( (LPVOID)&myStructure ))
			cout << "Found >" << pStructure->Data << "< myChar=" << pStructure->myChar 
				 << " Position=(" << myList.dlposn() << ")" << endl;		
	else    {
		cout << "Couldn't find it! - Status = " << myList.GetStatus() << endl;
	}
	

	/*----------------------------------------------------------------------
	 * Print the contents of the list so the user can believe it
	 *----------------------------------------------------------------------*/
	cout << "About to print the list again (forwards)" << endl;
	cout << "Press any key to continue" << endl;
	getch();
	cout << endl;

	pStructure = (Structure *)myList.dltofirst();
	while ( pStructure ) {
		cout	<< ">" << pStructure->Data << "< myChar=" << pStructure->myChar 
				<< " Position=(" << myList.dlposn() << ")" << endl;
		pStructure = ( Structure *)myList.dlgofwd();
	}

	/*----------------------------------------------------------------------
	 * Show the user the dlsetcompare() method
	 *----------------------------------------------------------------------*/
	cout << "About to change the comparison function using dlsetcompare()" << endl;
	cout << "sorting will also occur (different comparison function)" << endl;
	myList.dlsetcompare( Compare2 );
	myList.dlsort();	// Use the primitive swap sort algorithm

	cout << "Finished - About to print list contents (any key to continue)";
	getch();
	cout << endl;

	pStructure = (Structure *)myList.dltofirst();
	while ( pStructure ) {
		cout	<< ">" << pStructure->Data << "< myChar=" << pStructure->myChar 
				<< " Position=(" << myList.dlposn() << ")" << endl;
		pStructure = ( Structure *)myList.dlgofwd();
	}
	cout << "Note how the nodes are now sorted in DESCENDING order" << endl;

	/*-------------------------------------------------------
	 *         Now delete the 6th node in the list.         
	 *------------------------------------------------------*/

	cout << "About to delete NODE 6 (any key to continue)" << endl;
	getch();
	cout << endl;

	if ( pStructure = (Structure *)myList.dlfind( (LPVOID)&myStructure )) {
		myList.dldelete();
		cout << "DONE!.. about to print list contents" << endl;

		pStructure = (Structure *)myList.dltofirst();
		while ( pStructure ) {
			cout	<< ">" << pStructure->Data << "< myChar=" << pStructure->myChar 
					<< " Position=(" << myList.dlposn() << ")" << endl;
			pStructure = (Structure *)myList.dlgofwd();
		}
		cout << "There are now " << myList.dlcount() << " nodes in the list" << endl;
	}
	else {
		cout << "Couldn't find the 6th NODE in the list" << endl;
	}
	cout << "(any key to continue)";
	getch();
	cout << endl;

	/*-------------------------------------------------------
	 * Now insert a node in the list.  NOTE: If you want to
	 * insert a node at the beginning of the list use:    
	 * dlrewind THEN dlinsert, otherwise just use dlinsert.
	 *------------------------------------------------------*/
	cout << "About to insert a NODE at BEGINNING of LIST "
		"(Any key to continue)" << endl;
	getch();

	strcpy(myStructure.Data,"NODE=This is the NEWLY dlinserted one!");
	myStructure.myChar = new char[ kSmallBuf ];
	ASSERT( myStructure.myChar );
	strcpy(myStructure.myChar,"Inserted Number");
	myList.dlrewind();
	myList.dlinsert( ( LPVOID )&myStructure );

	pStructure = (Structure *)myList.dltofirst();
	cout << "\nAfter dlinsert......." << endl;
	while ( pStructure ) {
			cout	<< ">" << pStructure->Data << "< myChar=" << pStructure->myChar 
					<< " Position=(" << myList.dlposn() << ")" << endl;
			pStructure = (Structure *)myList.dlgofwd();
	}
	cout << "There are now " << myList.dlcount() << " nodes in the list" << endl;
	/*-------------------------------------------------------
	 * If you want to delete the nodes in the list, and wish
	 * to retain an empty list rather than delete it,
	 * then the following will suffice:               
	 *                                                     
	 *      pStructure =( Structure *)myList.dltofirst();                   
	 *      while ( pStructure ) {                                    
	 *              myList.dldelete();                          
	 *              pStructure = (Structure *)myList.dlget();               
	 *      }                                              
	 *                                                     
	 * The dlget function returns currency that has been set
	 * by the list Scanning pointer. On delete, the pointer 
	 * points to the next node in the list following the one
	 * that has been deleted, with the exception of the last
	 * node in the list. If this one is deleted, the Scanner
	 * pointer will point to the new last node in the list.
	 * dlget() returns a pointer to the data currently    
	 * pointed to by the Scanner pointer.		
	 *------------------------------------------------------*/

	/*-------------------------------------------------------
	 * Now test the dlreplace() method.
	 *------------------------------------------------------*/
	cout << endl;
	cout << "About to replace <DATA=10> with <DATA=11>" << endl;
	cout << "(any key to continue)";
	getch();

	strcpy( myStructure.Data,"DATA=10");
	cout << "Searching for DATA=10" << endl;
	if ( pStructure = (Structure *)myList.dlfind( (LPVOID)&myStructure )) {
		cout << "SUCCESSFUL find" << endl;
		// Insert some new data into our STRUCTURE 
		myStructure.myChar = new char[ kSmallBuf ];
		ASSERT( myStructure.myChar );
		strcpy(myStructure.Data,"DATA=11");
		strcpy(myStructure.myChar,"Now Number 11");
		myList.dlreplace( &myStructure );
	}
	else {
		cout << "Can't find the node in the list" << endl;
	}
	/*-------------------------------------------------------
	 * Print the contents of the list to show that it worked.
	 *------------------------------------------------------*/
	pStructure = ( Structure *)myList.dltofirst();
	while ( pStructure ) {
			cout	<< ">" << pStructure->Data << "< myChar=" << pStructure->myChar 
					<< " Position=(" << myList.dlposn() << ")" << endl;
			pStructure = (Structure *)myList.dlgofwd();
	}
	cout << "There are now " << myList.dlcount() << " nodes in the list" << endl;
	cout << "About to test list marking methods (any key to continue)";
	getch();
	cout << endl;
	/*-------------------------------------------------------
	 * 	Now test the list marking facilities.		
	 *------------------------------------------------------*/

	if (myList.dltomark()) {
		cout << "Should get an error here (Not marked)" << endl;
	}	
	if ( myList.dlatmark()) {
		cout << "Should get another error here (Not marked)" << endl;
	}
	cout << "(Any key to continue)" << endl;
	getch();

	/*-------------------------------------------------------
	 * Now let's mark the third node in the list.
	 *------------------------------------------------------*/
	cout << "About to mark the third node in the list" << endl;
	cout <<"(Any key to continue)";
	getch();

	strcpy(myStructure.Data,"DATA=03");

	if (pStructure =( Structure *)myList.dlfind( &myStructure )) {
		myList.dlmark();
		cout << "Moving away from marked node" << endl;
		pStructure = (Structure *)myList.dltolast();	// Move away from marked node
		cout << "Searching for previously marked node" << endl;
		myList.dltomark();
		if (myList.dlatmark()) {
			cout << "OK, at the marked node" << endl;
			/* Get a fresh copy of the data found */
			pStructure = (Structure *)myList.dlget();
			cout	<< ">" << pStructure->Data << "< myChar=" << pStructure->myChar 
					<< " Position=(" << myList.dlposn() << ")" << endl;
		}
	}
	else {
		cout << "Can't find the third node !!!!" << endl;
	}
	return( 0 );
}



/******************************************************************************
 *  Function Name: _tmain()
 *
 *  Purpose: The program's entry point.
 * 
 *  Parameters:    
 *			1.	argc - The number of command line arguments.
 *			2.	argv[] - The command line arguments themselves.
 *			3.	envp[] - The environment variables.
 *
 *  Side Effects:  None.
 *
 *  Return Values: The value of nRetCode.
 *
 ******************************************************************************/
int _tmain(int argc, TCHAR* argv[], TCHAR* envp[])
{
	int nRetCode = 0;

	// initialize MFC and print and error on failure
	if (!AfxWinInit(::GetModuleHandle(NULL), NULL, ::GetCommandLine(), 0))
	{
		// TODO: change error code to suit your needs
		cerr << _T("Fatal Error: MFC initialization failed") << endl;
		nRetCode = 1;
	}
	else
	{
		nRetCode = DemonstrateLinkedLists();
	}

	return nRetCode;
}


